function [planes, clusters, rest, restIndx] = extract_planes(fr, cam, minSize, method, display)
%% by Agnes Swadzba
% University Bielefeld
% Applied Informatics
% contact: aswadzba@techfak.uni-bielefeld.de
%
% [planes, clusters, rest, restIndx] = extract_planes(fr, cam, minSize, method, display)
% 
% frame: a struct consisting of distances, amplitudes, points3D, usage,
%        normals, ...
% cam: a struct for the camera consisting of focalLength, principalPoint,
%      pixelWidth, pixelHeight, ...
% minSize: specify the number points a plane has to consist of at least
% method: specify the method that will be used to form clusters of points within
%         the point cloud; 
%         possible: 
%           - 'regions' (cluster points lying inside of
%                        connected edges the socalled regions; a region image is
%                        realized through computing edges on a distances image)
%                        -> takes texture into consideration
%           - 'normals' (cluster points which have similar normals) 
%                        -> is not a very suitable method
%           - 'connectedPnts' (points are connected to their neighbors
%                              and therefore clustered into the same cluster, when the
%                              Euclidean distance in 3D between them is smaller than a
%                              certain threshold) 
%                              -> very good method
%           - 'planarClsts' (local planar patches are clustered to the
%                            same plane cluster depending on a conormality and a
%                            coplanarity measure) 
%                            -> very good method
%           - 'rGrow' is based on region growing using the information of
%                            local neighborhood. The method proposed by
%                            Olaf Kähler is used to realize a rough
%                            segmentation which is refined by RANSAC
%                            (e.g. extract two walls connected via a
%                            corner)

%% some default values:

if(nargin < 5)
    display = false;
end

if(strcmp(method, 'histNormals') || strcmp(method, 'histNormals+gmd'))
    pntsNew = position_point_cloud(fr.points3D);
    fr.points3D = pntsNew;
end

rest = [];
restIndx = [];

%% choose and cluster points of the point cloud using a specified method
disp('Initial clustering');

if(strcmp(method, 'regions'))
    [clusters, clusterIndx, rest, restIndx] = cluster_through_regions(fr, cam);
elseif(strcmp(method, 'normals'))
    % threshold = 30° = 3*pi/18 = 0.5236
    [clusters, clusterIndx] = cluster_through_normals(fr, cam, 0.5236);
elseif(strcmp(method, 'connectedPnts'))
    % camProportion = 0.04 / 8;
    camProportion = 1;
    [clusters, rest, clusterIndx] = cluster_through_connectedPoints(fr);
    %clusters = cluster_through_connectedPoints_rgrow(pnts, width, height, camProportion, usage);
elseif(strcmp(method, 'planarClsts'))
    [clusters, rest, clusterIndx, restIndx] = pnts2plane_classification_rgrow(fr, cam);
elseif(strcmp(method, 'rGrow'))
    [clusters, rest, clusterIndx, restIndx] = pnts2plane_classification_rgrowpure(fr, cam);
elseif(strcmp(method, 'gmd'))
    restIndx = 1:cam.width*cam.height;
    restIndx(fr.usage) = [];
    rest = fr.points3D(1:3,restIndx);
    [clusters, clusterIndx] = cluster_through_gmd(fr.points3D(1:3,fr.usage));
    clusterIndx = cellfun(@(x)(fr.usage(x)), clusterIndx, 'UniformOutput', false);
elseif(strcmp(method, 'histNormals'))
    rest = [];
    restIndx = [];
    [clusters, clusterIndx] = cluster_through_histNormals(fr, cam);
elseif(strcmp(method, 'histNormals+gmd'))
    rest = [];
    restIndx = [];
    [cl, clIndx] = cluster_through_histNormals(fr, cam);
    
    clusters = [];
    clusterIndx = [];
    i = 1;
    while(i <= length(cl))
        [help, indx] = cluster_through_gmd(cl{i});
        orgIndx = clIndx{i};
        helpIndx = cellfun(@(x)(orgIndx(x)), indx, 'UniformOutput', false);
        clusters = [clusters help];
        clusterIndx = [clusterIndx helpIndx];
        
        i = i+1;
    end
elseif(strcmp(method, 'cuefusion'))
    [cl1, cl1Indx] = cluster_through_histNormals(fr, cam);
    [cl2, cl2Indx] = cluster_through_regions(fr, cam);
    
    restIndx = 1:cam.width*cam.height;
    
    clIndx = [];
    i = 1;
    while( i <= length(cl1Indx) )
        
        j = 1;
        while( j <= length(cl2Indx) )
        
            interIndx = intersect(cl1Indx{i}, cl2Indx{j});
            clIndx = [clIndx {interIndx}];
            
            restIndx(interIndx) = 0;
            
            j = j+1;
            
        end
        
        i = i+1;
        
    end
    
    clusterIndx = clIndx;
    clusters = [];
    
    i = 1;
    while( i <= length(clIndx) )
        
        clusters = [clusters {fr.points3D(1:3, clIndx{i})}];
        
        i = i+1;
        
    end
    
    restIndx(find(restIndx == 0)) = [];
    restIndx = intersect(restIndx, fr.usage);
    rest = fr.points3D(1:3, restIndx);
    
    show_clusters(clusters);
    
else
    disp('Error: Unknown clustering method! Please choose between the regions or the normals method');
end

if(display)
    fig = figure;
    fig = show_clusters(clusters, 0, fig, [], [], 1);
    if(~isempty(rest))
            fig = plot_3Dpoints(rest, 'k', fig, 0.3);
    end
end

inCl = clusters;
inClIndx = clusterIndx;
l = cellfun('size', clusters, 2);
lo = find(l < minSize);
rest = [rest cat(2, clusters{lo})];
restIndx = [restIndx cat(2, clusterIndx{lo})];
inCl(lo) = [];
inClIndx(lo) = [];
planesIndx = inClIndx;

%% refining the extracted planar cluster using RANSAC
disp('Refining the extracted planar cluster using RANSAC');

planes = [];
i = 1;
while( i <= length(inCl))
    
    if(isempty(planesIndx{i}))
        i = i+1;
        
        continue;
    end
    
    frame = fr;
    frame.usage = planesIndx{i};
    
    %[p, r, rIndx] = plane_segmentation(inCl{i},planesIndx{i});
    [cl, clIndx, p, r, rIndx] = pnts2plane_classification_RANSAC(frame, cam); 
    if(isempty(p))
        rest = [rest r];
        restIndx = [restIndx rIndx];
        i = i+1;
        
        continue;
    end
    
%     points = inCl{i};
%     indx = planesIndx{i};
%     [posEnd, nEnd, dEnd] = main_plane_extraction(points);    
%     plane.pnts = points(1:3, posEnd);
%     plane.n = nEnd;
%     plane.d = dEnd;
%     plane.indx = indx(posEnd);
%     planes = [planes plane];
%    
%     points(:, posEnd) = [];
%     rest = [rest points];
    
    l = cellfun('size', {p.pnts}, 2);
    li = find(l >= minSize);
    lo = find(l < minSize);
    
    planes = [planes p(li)];
    rest = [rest r [p(lo).pnts]];
    restIndx = [restIndx rIndx [p(lo).indx]];
     
    i = i+1;
    
end

%% assign plane-id
i=1;
while( i<= length(planes) ) planes(i).id = i; i = i+1; end

%% displaying planes
if(display)
    fig = figure;
    fig = show_clusters({planes.pnts}, 0, fig, [], [], 1);
    if(~isempty(rest))
        fig = plot_3Dpoints(rest, 'k', fig, 0.3);
    end
end

