function [planes, rest] = pnts2plane_classification(pnts, width, height, usage, display)
% [planes, rest] = pnts2plane_classification(pnts, width, height, usage, display)
%
%   This function classifies points as planar or not. Afterwards the planar
%   points are associated with a certain plane. The function is based on the
%   paper "Geometry and Texture Recovery" of Stamos and Allen. This function
%   does not take into account the neighborhood (based on the organization
%   within the matrix structure) of each point. No use of region growing
%   techniques.
%
% pnts: set of points with the following structure
%           | p1x p2x p3x p4x     |
%    pnts = | p1y p2y p3y p4y ... |
%           | p1z p2z p3z p4z     |
% width, height: dimension of the matrix the points of 'pnts' are organized in
% usage: specifies which of the points in 'pnts' are valid points
% display: 1 - visualize clusters; 0 - no visulaization (optinal; default: 0)

if(nargin < 5)
    display = false;
end

if(nargin < 4)
    usage = 1:1:size(pnts,2);
end

rest = [];

% compute normals for each point of pnts
[normals, d, deviation] = local_plane_patches(pnts, width, height, 5);

% compute the deviation of each point to its corresponding plane
dev = zeros(1, size(pnts, 2));
for j = 1:size(pnts, 2)
    dev(j) = normals(:,j)'*pnts(:,j) - d(j);
end

% classify if points are planar or not:
indxPlanar = find(deviation <= 2000);
indxNotPlanar = find(deviation > 2000);

indxPlanar = indxPlanar(ismember(indxPlanar, usage));
indxNotPlanar = indxNotPlanar(ismember(indxNotPlanar, usage));

planarPnts = pnts(:, indxPlanar);
rest = [rest pnts(:, indxNotPlanar)];
planarNormals = normals(:, indxPlanar);
planarDev = dev(indxPlanar);

planarPnts = planarPnts + [planarDev.*planarNormals(1,:); ... 
    planarDev.*planarNormals(2,:);planarDev.*planarNormals(3,:)];

patches = [];
i = 1;
while(~isempty(planarPnts))
    
    workingSet = planarPnts(:,1);
    workingNormals = planarNormals(:,1);
    
    planarPnts(:,1) = [];
    planarNormals(:,1) = [];
    
    patches{i} = [];
    
    while(~isempty(workingSet))
        
        currPnt = workingSet(:,1);
        currN = workingNormals(:,1);
        
        workingSet(:,1) = [];
        workingNormals(:,1) = [];
        
        patches{i} = [patches{i} currPnt];
        
        % connect points with via conormality measure: (similar angles between
        % normals)
        angles = acos(currN'*planarNormals);
        
        in = find(angles < (1/18 * pi));
        
        workingSet = [workingSet planarPnts(:, in)];
        workingNormals = [workingNormals planarNormals(:, in)];
        
        planarPnts(:, in) = [];
        planarNormals(:, in) = [];
        
        if(isempty(workingSet)) continue; end
        
        % select points which fulfill the coplanarity measure:
        % patches lie on the same infinite plane and a near enough to each other
        disVec = workingSet - currPnt*ones(1, size(workingSet, 2));
        
        firstTerm = abs(currN'*disVec);
        secondTerm = zeros(1, size(disVec, 2));
        for j = 1:size(disVec, 2)
            secondTerm(j) = abs(workingNormals(:,j)'*disVec(:,j));
        end
        
        decider = max([firstTerm; secondTerm]);
        
        out = find(decider >= 100);
        
        planarPnts = [planarPnts workingSet(:, out)];
        planarNormals = [planarNormals workingNormals(:, out)];
        workingSet(:, out) = [];
        workingNormals(:, out) = [];
        
    end
    
    i = i+1;
    
end
        
if(display)
    fig = show_clusters(patches, 50);
    fig = plot_3Dpoints(rest, 'k', fig);
end
