function [planes, rest, patchesIndx, restIndx] = pnts2plane_classification_rgrow(fr, cam, display)
%% [planes, rest, patchesIndx, restIndx] = pnts2plane_classification_rgrow(fr, cam, display)
%
%   This function classifies points as planar or not. Afterwards the planar
%   points are associated with a certain plane. The function is based on the
%   paper "Geometry and Texture Recovery" of Stamos and Allen. This function
%   does take into account the neighborhood (based on the organization
%   within the matrix structure) of each point. Use of region growing
%   techniques.
%
% pnts: set of points with the following structure
%           | p1x p2x p3x p4x     |
%    pnts = | p1y p2y p3y p4y ... |
%           | p1z p2z p3z p4z     |
% width, height: dimension of the matrix the points of 'pnts' are organized in
% usage: specifies which of the points in 'pnts' are valid points
% display: 1 - visualize clusters; 0 - no visulaization (optinal; default: 0)

%% some default values:

if(nargin < 3)
    display = false;
end

if(~isfield(fr, 'normals'))
    seq.frames = fr;
    seq.camType = cam;
    
    seq = preprocessing(seq, 'normals');
    fr = seq.frames;
end

rest = [];
restIndx = [];

normals = fr.normals;
%deviation = fr.dev;
usage = fr.usage;
pnts = fr.points3D;
planarIndx = fr.planar;
planar = fr.planar;
restIndx = fr.notplanar;
rest = pnts(1:3, restIndx);

%% start region growing:

patchesIndx = [];
patches = [];
i = 1;
while(~isempty(planarIndx))
    
    workingIndx =  planarIndx(1);
    
    planarIndx(1) = [];
    
    patchesIndx{i} = [];
    
    while(~isempty(workingIndx))
        
        currIndx = workingIndx(1);
        workingIndx(1) = [];
        patchesIndx{i} = [patchesIndx{i} currIndx];
        
        % find indices of neighbors
        %ngbIndx = getNgbIndx(currIndx, width, height, 3); %Swissranger: 8-neighboorhood
        %ngbIndx = getNgbIndx(currIndx, width, height, 21); %DC_Tracking: 440-neighboorhood
        ngbIndx = fr.indx8{currIndx};
        
        % find the indices, that are unprocessed until now and refer to a planar
        % point (-> stored in 'planar')
        %ngbIndx = ngbIndx(ismember(ngbIndx, usage));
        ngbIndx = ngbIndx(ismember(ngbIndx, planar));
        
        if(isempty(ngbIndx)) continue; end
        
        % conormality measure:
        currNormal = normals(:, currIndx);
        ngbNormals = normals(:, ngbIndx);
        
        angles = acos(currNormal'*ngbNormals);
        % find acute angles
        f = find(angles > pi/2);
        angles(f) = pi - angles(f);
        in = find(angles < (1/18 * pi));
        
        ngbIndx = ngbIndx(in);
        
        if(isempty(ngbIndx)) continue; end
        
        % coplanarity measure:
         currPnt = pnts(1:3, currIndx);
         ngbPnts = pnts(1:3, ngbIndx);
         ngbNormals = normals(:, ngbIndx);
         
         r12 = ngbPnts - currPnt * ones(1, size(ngbPnts, 2));
         
         firstTerm = abs(currNormal'*r12);
         secondTerm = zeros(1, size(r12, 2));
         for j = 1:size(r12, 2)
             secondTerm(j) = abs(ngbNormals(:,j)'*r12(:,j));
         end
         
         decider = max([firstTerm; secondTerm]);
         %threshold = log(10*currPnt(3));
         %threshold = 32 * 0.05 * currPnt(3); %Swissranger
         threshold = 4 * 0.05 * currPnt(3); %Swissranger
         %threshold = 100; %DC_Tracking
         %threshold = 30;
         in = find(decider < threshold);
         
         ngbIndx = ngbIndx(in);
        
        workingIndx = [workingIndx ngbIndx];
        workingIndx = unique(workingIndx);
        workingIndx(ismember(workingIndx, patchesIndx{i})) = [];
        
    end
    
    patches{i} = pnts(1:3, patchesIndx{i});
    planarIndx(ismember(planarIndx, patchesIndx{i})) = [];
    
    i = i+1;
    
end
    
%% display results:

if(display)
    fig = show_clusters(patches, 50);
    fig = plot_3Dpoints(rest, 'k', fig);
end

planes = patches;
